/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Helper routines for the pattern mutators.
 */


#include "defs.h"


/* buffer for storing temporary values. Declared globally to conserve stack
 * space and prevent having to keep dynamically allocating memory whenever
 * some scratch space is needed.
 */
int temp_i[TEMP_I_SIZE];

int temp_i_size = 0;



/* fill_temp_with_pitches:
 *  Fills the temporary buffer with a list of all the pitches used in a
 *  pattern, sorting them into ascending order.
 */
void fill_temp_with_pitches(p)
PATTERN *p;
{
    int c;
    int x, x2;

    temp_i_size = 0;

    if (!p)
	return;

    /* insert pitches into temp_i, sorting them into ascending order */ 
    for (c=0; (c<p->length) && (temp_i_size<TEMP_I_SIZE); c++) {
	for (x=0; x<c; x++)
	    if (temp_i[x] > p->note[c].pitch)
		break;

	for (x2=temp_i_size; x2>x; x2--)
	    temp_i[x2] = temp_i[x2-1];

	temp_i[x] = p->note[c].pitch;
	temp_i_size++;
    }
}



/* fill_temp_with_unique_pitches:
 *  Fills the temporary buffer with a list of all the pitches used in a
 *  pattern, removing duplicates of the same pitch.
 */
void fill_temp_with_unique_pitches(p)
PATTERN *p;
{
    int c;
    int x, x2;

    temp_i_size = 0;

    if (!p)
	return;

    /* insert pitches into temp_i, sorting them into ascending order */ 
    for (c=0; (c<p->length) && (temp_i_size<TEMP_I_SIZE); c++) {
	for (x=0; x<temp_i_size; x++) {
	    if (temp_i[x] == p->note[c].pitch)
		goto skip_duplicate;

	    if (temp_i[x] > p->note[c].pitch)
		break;
	}

	for (x2=temp_i_size; x2>x; x2--)
	    temp_i[x2] = temp_i[x2-1];

	temp_i[x] = p->note[c].pitch;
	temp_i_size++;

	skip_duplicate:
	    ;
    }
}



/* fill_temp_with_mode:
 *  Fills the temporary buffer with a list of all the pitches used in the
 *  phrase, reduced to a single octave so that a mode can be extracted from
 *  it.
 */
void fill_temp_with_mode(p)
PATTERN *p;
{
    int c;

    for (c=0; c<12; c++)
	temp_i[c] = FALSE;

    if (!p)
	return;

    for (c=0; c<p->length; c++)
	temp_i[p->note[c].pitch % 12] = TRUE;
}



/* closest_note_in_mode:
 *  Using the mode stored in the temporary buffer by fill_temp_with_mode(),
 *  rounds i to the closest pitch in the mode.
 */
int closest_note_in_mode(i)
int i;
{
    int c;

    for (c=0; c<7; c++) {
	/* look for lower pitches */
	if (temp_i[(i-c)%12])
	    return i-c;

	/* and look for higher pitches */
	if (temp_i[(i+c)%12])
	    return i+c;
    }

    /* if the mode is empty, better just return i */
    return i;
}



/* median_pitch:
 *  Returns the average (median) of the pitches used in a pattern.
 */
int median_pitch(p)
PATTERN *p;
{
    fill_temp_with_pitches(p);

    return temp_i[temp_i_size/2];
}



/* clone_pattern:
 *  Creates and returns an exact copy of an existing pattern.
 */
PATTERN *clone_pattern(p)
PATTERN *p;
{
    PATTERN *np;
    int x;

    if (!p)
	return NULL;

    np = malloc((int)(sizeof(PATTERN) + p->length * sizeof(NOTE)));

    if (np) {
	np->channel = p->channel;
	np->length = np->allocated_size = p->length; 

	for (x=0; x<p->length; x++)
	    np->note[x] = p->note[x];
    }

    return np;
}

