/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Routines for working with linked lists, and picking items from lists
 *  using a weighted probability system.
 */


#include "defs.h"



/* add_to_list:
 *  Adds an item to the specified list.
 */
void add_to_list(list, item, coolness)
LIST_MEMBER *list;
long item;
long coolness;
{
    if ((!list) || (!item))
	return;

    while (list->data)      /* find a free slot */
	list++;

    list->desc = NULL;
    list->data = item;
    list->coolness = coolness;
}



/* remove_from_list:
 *  Removes an item from a list.
 */
void remove_from_list(list, item)
LIST_MEMBER *list;
long item;
{
    if ((!list) || (!item))
	return;

    while (list->data != item) {        /* find the item */
	if (list->data == 0)
	    return;

	list++;
    }

    while (list->data) {                /* shift others back to fill hole */
	list[0] = list[1];
	list++;
    }
}



/* get_list_size:
 *  Counts the number of items in a list.
 */
int get_list_size(list)
LIST_MEMBER *list;
{
    int c = 0;

    if (!list)
	return 0;

    while (list->data) {
	c++;
	list++;
    }

    return c;
}



/* pick_from_list:
 *  Uses item weightings to randomly select an item from a list. The higher 
 *  the weighting of an entry, the more likely it is to be selected. The
 *  coolness parameter will be set to the probablility weighting of the
 *  selected item.
 */
long pick_from_list(list, coolness)
LIST_MEMBER *list;
long *coolness;
{
    long total;
    long val;
    LIST_MEMBER *p;

    if (!list)
	return 0;

    total = 0;
    p = list;

    /* count the total probablities */
    while (p->data) {
	total += p->coolness;
	p++;
    }

    /* bottle out if there is a zero chance of picking anything... */
    if (total <= 0) {
	*coolness = 0;
	return 0;
    }

    /* make a linear selection from the total probablity range */
    val = random() % total;

    /* find the item at the selected spot, using the probablity weights */
    p = list;
    while (p->data) {
	val -= p->coolness;
	if (val < 0) {
	    *coolness = p->coolness;
	    return p->data;
	}
	p++; 
    }

    /* should never reach this point! */ 
    *coolness = 0;
    return 0;
}



/* pick_bad_item:
 *  Picks the item with the lowest weighting in the list.
 */
long pick_bad_item(list)
LIST_MEMBER *list;
{
    int pos;
    int best = MAX_PROB;
    long best_data = 0;

    if (!list)
	return 0;

    for (pos=0; list[pos].data; pos++) {
	if (list[pos].coolness < best) {
	    best = list[pos].coolness;
	    best_data = list[pos].data;
	}
    }

    return best_data;
}



/* adjust_weighting:
 *  Adjusts the probability weighting of an entry in a list.
 */
void adjust_weighting(list, item, adjustment)
LIST_MEMBER *list;
long item;
long adjustment;
{
    if ((!list) || (!item))
	return;

    /* need to redraw the user interface? */
    if (list == view_list)
	ui_update = TRUE;

    /* find and adjust the item */
    while (list->data) {
	if (list->data == item) {
	    list->coolness += adjustment;

	    if (list->coolness < 1)
		list->coolness = 1;
	    else if (list->coolness > MAX_PROB)
		list->coolness = MAX_PROB;

	    return;
	}

	list++; 
    }
}

