/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Low level MIDI recording and playback. The core routine is do_midi(), 
 *  which should be called as often as possible. This carries out all
 *  the neccessary playback, recording, and MIDI thru operations.
 */


#include "defs.h"


/* MIDI output file */
char *log_name = NULL;
FILE *log_file = NULL;
int log_len = 0;
long log_time = 0;
long log_ticks = 0;

/* if non-zero, the metronome is turned on */
int metronome_flag = TRUE;

/* tempo in bpm */
int metronome_speed = 120;

/* General MIDI drum channel and hi-hat sound */
int metronome_channel = 9;
int metronome_note = 42;

/* the active channel for MIDI record/thru */
int midi_channel = 0;

/* current position of the pitch bend wheel */
long pitch_bend = 0;

/* if non-zero, recorded data will be quantized */
int quantize_flag = TRUE;

/* the fraction of a beat to quantize to (4 = quarter beats) */
int quantize_speed = 4;

/* internal counters for the metronome */
long metronome_time = 0;
int metronome_on = FALSE;

/* set this to start recording on the next downbeat */
int start_record = FALSE;

/* for storing MIDI data as it is coming in */
unsigned char buffer[3];
int buffer_size = 0;
int message_size = 0;
int in_sysex = FALSE;


/* table to find the size of each MIDI message */
int size_of_message[] = {
    0, 0, 0, 0, 0, 0, 0, 0,
    3,  /* note off */
    3,  /* note on */
    3,  /* note aftertouch */
    3,  /* control change */
    2,  /* program change */
    2,  /* channel aftertouch */
    3,  /* pitch bender */
    1   /* real time events */
};


/* somewhere to put data while we are recording */
PLAYING_PATTERN rec = { NULL, 0, 0 };


/* when a note-on message is recieved during a record, there is no way of 
 * telling how long it will last. This array stores the position within the
 * phrase being recorded of all the notes which have been turned on but not 
 * yet stopped, so we can go back and fill in the length field when a 
 * note-off is recieved.
 */
int active_note[128];


/* this is a hack to fix an irritating problem. When recording short patterns,
 * I tend to hit the sustain pedal at the same time as playing the first
 * note of the pattern. If the note-on arrives before the pedal controller,
 * the first notes will be missed out. This array stores all the notes that
 * are currently active (-1 means note not playing, positive values are the
 * velocity), so that when entering record mode we can catch up with any
 * notes that are already sounding.
 */
int note_status[128];


/* this list is used by the playback code: when it triggers a note it adds
 * an entry to this list, which will cause the note to be turned off at the
 * appropriate time.
 */

#define MAX_TURN_OFF    1024

struct
{
    unsigned char channel;
    unsigned char pitch;
    long time;
} turn_off[MAX_TURN_OFF];

int turn_off_count = 0;



/* midi_puts:
 *  Writes a string to the MIDI logging file.
 */
void midi_puts(s)
char *s;
{
    if (log_file) {
	while (*s) {
	    fputc(*s, log_file);
	    s++;
	    log_len++;
	}
    }
}



/* midi_putl:
 *  Writes a long to the MIDI logging file.
 */
void midi_putl(l)
long l;
{
    int b1, b2, b3, b4;

    if (log_file) {
	b1 = (int)((l & 0xFF000000L) >> 24);
	b2 = (int)((l & 0x00FF0000L) >> 16);
	b3 = (int)((l & 0x0000FF00L) >> 8);
	b4 = (int)l & 0xFF;

	fputc(b1, log_file);
	fputc(b2, log_file);
	fputc(b3, log_file);
	fputc(b4, log_file);

	log_len += 4;
    }
}



/* midi_putw:
 *  Writes a word to the MIDI logging file.
 */
void midi_putw(w)
short w;
{
    int b1, b2;

    if (log_file) {
	b1 = (w & 0xFF00) >> 8;
	b2 = w & 0xFF;

	fputc(b1, log_file);
	fputc(b2, log_file);

	log_len += 2;
    }
}



/* midi_putc:
 *  Writes a byte to the MIDI logging file.
 */
void midi_putc(c)
char c;
{
    if (log_file) {
	fputc(c, log_file);
	log_len++;
    }
}



/* midi_put_time:
 *  Writes a time delta into the MIDI logging file.
 */
void midi_put_time()
{
    long gap, t;
    int extra, got;

    if ((log_file) && (metronome_time > 0)) {
	gap = (CLK_TCK * 60) / metronome_speed;
	t = my_clock - metronome_time;

	extra = 0;

	while (t >= gap) {
	    extra++;
	    t -= gap;
	}

	t = t*120/gap + (log_ticks+extra)*120 - log_time;

	got = FALSE;

	if ((t>>21)&0x7F) {
	    fputc(((t>>21)&0x7F)|0x80, log_file);
	    log_len++;
	    got = TRUE;
	}

	if ((got) || ((t>>14)&0x7F)) {
	    fputc(((t>>14)&0x7F)|0x80, log_file);
	    log_len++;
	    got = TRUE;
	}

	if ((got) || ((t>>7)&0x7F)) {
	    fputc(((t>>7)&0x7F)|0x80, log_file);
	    log_len++;
	}

	fputc(t&0x7F, log_file);
	log_len++;

	log_time += t;
    }
}



/* note_off:
 *  Handles a note off event which has been recieved from the MIDI input,
 *  passing it through to MIDI out and recording it if required.
 */
void note_off(note)
int note;
{
    NOTE *n;

    midi_output(0x80+midi_channel);
    midi_output(note);
    midi_output(0);

    midi_put_time();
    midi_putc(0x80+midi_channel);
    midi_putc(note);
    midi_putc(0);

    if ((rec.pattern) && (active_note[note] >= 0)) {
	n = rec.pattern->note + active_note[note];
	n->length = my_clock - (rec.start_time + n->time);
	active_note[note] = -1;
    }

    note_status[note] = -1;
}



/* record_note:
 *  Adds a note to the pattern currently being recorded
 */
void record_note(note, vel)
int note, vel;
{
    NOTE *n;
    PATTERN *p;

    p = resize_pattern(rec.pattern, rec.pattern->length+1);
    if (!p) {
	stop_recording();
	return;
    }

    rec.pattern = p;

    n = p->note + p->length;

    n->pitch = note;
    n->vel = vel;
    n->time = my_clock - rec.start_time;
    n->length = 0;  /* unknown at the moment */

    if (active_note[note] >= 0)
	note_off(note);

    active_note[note] = p->length;

    p->length++;
}



/* note_on:
 *  Handles a note on event recieved from the MIDI input, passing it through
 *  to MIDI out and recording it as required.
 */
void note_on(note, vel)
int note, vel;
{
    midi_output(0x90+midi_channel);
    midi_output(note);
    midi_output(vel);

    midi_put_time();
    midi_putc(0x90+midi_channel);
    midi_putc(note);
    midi_putc(vel);

    if (rec.pattern)
	record_note(note, vel);

    note_status[note] = vel; 
}



/* start_recording:
 *  Start recording a new pattern.
 */
void start_recording()
{
    int x;

    TRACE("start_recording()\n");

    if (rec.pattern)     /* quit if already recording */
	return;

    for (x=0; x<128; x++)
	active_note[x] = -1;

    rec.pattern = malloc((int)sizeof(PATTERN)); 
    if (!rec.pattern)
	return;

    rec.pattern->channel = midi_channel;
    rec.pattern->length = rec.pattern->allocated_size = 0;
    rec.start_time = my_clock;

    for (x=0; x<128; x++)
	if (note_status[x] >= 0)
	    record_note(x, note_status[x]);
}



/* quantize:
 *  Quantizes a recorded pattern to prevent rhythmic innaccuracies.
 */
void quantize(pattern)
PATTERN *pattern;
{
    int x;
    long gap = (CLK_TCK * 60) / metronome_speed;
    long qgap = gap / quantize_speed;
    long beat;
    long rem;

    if ((quantize_flag) && (pattern)) {
	for (x=0; x<pattern->length; x++) {
	    /* complicated method to prevent rounding errors */
	    beat = pattern->note[x].time / gap;
	    rem = pattern->note[x].time - (beat * gap); 
	    rem = ((rem + qgap/2) / qgap) * qgap;
	    pattern->note[x].time = (beat * gap) + rem;
	} 
    }
}



/* stop_recording:
 *  Stops recording a pattern when the sustain pedal is released.
 */
void stop_recording()
{
    int x;

    TRACE("stop_recording()\n");

    if (!rec.pattern)
	return;

    for (x=0; x<128; x++)
	if (active_note[x] >= 0)
	    note_off(x);

    quantize(rec.pattern);

    rec.pattern->parent = NULL;
    rec.pattern->mutator = NULL;
    rec.pattern->mutator_i = 0;

    add_pattern(rec.pattern, MAX_PROB/4);

    rec.pattern = NULL;
}



/* is_recording:
 *  Checks if we are currently recording something.
 */
int is_recording()
{
    return (rec.pattern != NULL);
}



/* deal_with_event:
 *  Processes a MIDI input event once it has been totally recieved.
 */
void deal_with_event()
{
    switch (buffer[0] & 0xF0) {

	case 0x80:
	    note_off(buffer[1]);
	    break;

	case 0x90:
	    if (buffer[2] == 0)
		note_off(buffer[1]);
	    else
		note_on(buffer[1], buffer[2]);
	    break;

	case 0xB0:
	    if (buffer[1] == 0x40) {
		if (buffer[2] < 0x40)
		    stop_recording();
		else
		    start_record = TRUE;  /* flag to start on next downbeat */
	    }
	    break;

	case 0xE0:
	    pitch_bend = (((long)buffer[2]<<7) + buffer[1]) - 8192;
	    pitch_bend /= 64;
	    break;
    }

    buffer_size = message_size = 0;
}



/* process_input:
 *  Deals with any waiting midi input.
 */
void process_input()
{
    unsigned char b;

    while (midi_waiting()) {
	b = midi_in();

	/* skip system exclusive data */
	if (in_sysex) {
	    if (b == 0xF7)
		in_sysex = FALSE;
	}
	else {
	    /* got all the bytes of an event? */
	    if ((buffer_size < message_size) && (!(b & 0x80))) {
		buffer[buffer_size++] = b;
		if (buffer_size == message_size)
		    deal_with_event();
	    }
	    else {
		/* sysex or active sensing data? */
		if ((b & 0xF0) == 0xF0) {
		    if (b == 0xF0)
			in_sysex = TRUE;
		}
		else {
		    /* start of a new event */
		    if (b & 0x80) {
			buffer[0] = b;
			buffer_size = 1;
		    }
		    else {      /* running status */
			buffer[1] = b;
			buffer_size = 2;
		    }
		    message_size = size_of_message[buffer[0]>>4]; 
		} 
	    }
	}
    }
}



/* playback:
 *  MIDI pattern playback controller.
 */
void playback()
{
    long time = my_clock;
    long off_time;
    unsigned char channel, pitch, vel;
    int x, y, z;
    PLAYING_PATTERN *p;

    for (x=0; x<MAX_PATTERNS; x++) {
	p = playing_pattern+x;
	if (p->pattern) {
	    while ((p->pattern->note[p->pos].time + p->start_time) <= time) {
		channel = p->pattern->channel;
		pitch = p->pattern->note[p->pos].pitch;
		vel = p->pattern->note[p->pos].vel;
		off_time = time + p->pattern->note[p->pos].length;

		if (turn_off_count >= MAX_TURN_OFF)
		    break;

		midi_output(0x90+channel);
		midi_output(pitch);
		midi_output(vel);

		midi_put_time();
		midi_putc(0x90+channel);
		midi_putc(pitch);
		midi_putc(vel);

		for (y=0; y<turn_off_count; y++)
		    if (turn_off[y].time > off_time)
			break;

		for (z=turn_off_count; z>y; z--)
		    turn_off[z] = turn_off[z-1];

		turn_off_count++;
		turn_off[y].channel = channel;
		turn_off[y].pitch = pitch;
		turn_off[y].time = off_time; 

		p->pos++;
		if (p->pos >= p->pattern->length) {
		    p->pattern = NULL;
		    p->pos = 0;
		    p->start_time = 0;
		    break;
		}
	    }
	}
    }

    while ((turn_off_count > 0) && (time >= turn_off[0].time)) {
	midi_output(0x80+turn_off[0].channel);
	midi_output(turn_off[0].pitch);
	midi_output(0);

	midi_put_time();
	midi_putc(0x80+turn_off[0].channel);
	midi_putc(turn_off[0].pitch);
	midi_putc(0);

	turn_off_count--;
	for (x=0; x<turn_off_count; x++)
	    turn_off[x] = turn_off[x+1];
    }
}



/* do_metronome:
 *  Processes metronome ticks.
 */
void do_metronome()
{
    long t;
    long gap = (CLK_TCK * 60) / metronome_speed;

    if (metronome_time <= 0)
	metronome_time = my_clock;

    t = my_clock - metronome_time;

    while (t >= gap) {
	t -= gap;
	metronome_time += gap;
	log_ticks++;

	/* turn off the old note */
	if (metronome_on) {
	    midi_output(0x80 + metronome_channel);
	    midi_output(metronome_note);
	    midi_output(0);

	    midi_put_time();
	    midi_putc(0x80 + metronome_channel);
	    midi_putc(metronome_note);
	    midi_putc(0);
	}

	/* turn on new note */
	if (metronome_flag) {
	    midi_output(0x90 + metronome_channel);
	    midi_output(metronome_note);
	    midi_output(100);

	    midi_put_time();
	    midi_putc(0x90 + metronome_channel);
	    midi_putc(metronome_note);
	    midi_putc(100);

	    metronome_on = TRUE; 
	}
	else
	    metronome_on = FALSE;

	if (start_record) {
	    start_recording();
	    start_record = FALSE;
	}

	play_patterns();
    }
}



/* do_midi:
 *  Main MIDI controller routine: call as often as possible.
 */
void do_midi()
{
    my_clock = clock();
    process_input(); 
    playback();
    do_metronome();
}



/* midi_init:
 *  Sets up the MIDI code ready for use.
 */
void midi_init()
{
    int x;

    for (x=0; x<128; x++) {
	active_note[x] = -1;
	note_status[x] = -1;
    }

    while (midi_waiting())
	midi_in();
}



/* all_notes_off:
 *  Sends all-notes-off messages on every channel
 */
void all_notes_off()
{
    int x;

    for (x=0; x<16; x++) {
	midi_output(0xB0+x);
	midi_output(0x7B);
	midi_output(0);
    }
}



/* midi_cleanup:
 *  Shuts down the midi player at program termination.
 */
void midi_cleanup()
{
    if (rec.pattern != NULL)
	stop_recording();

    all_notes_off();
}



/* adjust_times:
 *  Adjusts all the stored times after a pause.
 */
void adjust_times(t)
long t;
{
    int x;

    for (x=0; x<MAX_PATTERNS; x++)
	playing_pattern[x].start_time += t;

    rec.start_time += t;
    metronome_time += t;
}



/* change_midi_channel:
 *  Called by the user interface when the midi channel is altered, to
 *  turn off any active notes.
 */
void change_midi_channel(c)
int c;
{
    int x;

    for (x=0; x<128; x++)
	if (note_status[x] >= 0)
	    note_off(x);

    midi_channel = c;
}

