/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Mutation routines for making alterations to the rhythm of a pattern.
 */


#include "defs.h"


MUTATE_FUNC stretcher_func, copy_rhythm_func, cut_block_func, 
	    paste_blocks_func;


/* different types of stretching */

#define AUGMENTATION        1
#define DIMINUTION          2


LIST_MEMBER stretcher_params[] =
{
    /* <description>                <data>                        <coolness> */
    { "Augmentation",               AUGMENTATION,                 MAX_PROB/4 },
    { "Diminution",                 DIMINUTION,                   MAX_PROB/4 },
    { NULL, 0, 0 }
};



MUTATOR stretcher =
{
    stretcher_func,             /* stretches a pattern */
    stretcher_params            /* by how much? */
};



MUTATOR copy_rhythm =
{
    copy_rhythm_func,           /* copies a rhythm from a different pattern */
    NULL                        /* don't need an integer parameter */
};



MUTATOR cut_block =
{
    cut_block_func,             /* cuts a small block out of a pattern */
    NULL                        /* don't need an integer parameter */
};



/* different ways to paste blocks together */

#define PASTE_AT_PITCH      1
#define PASTE_TRANSPOSED    2
#define PASTE_MODAL         3


LIST_MEMBER paste_blocks_params[] =
{
    /* <description>                <data>                        <coolness> */
    { "Paste at Pitch",             PASTE_AT_PITCH,               MAX_PROB/4 },
    { "Paste Transposed",           PASTE_TRANSPOSED,             MAX_PROB/4 },
    { "Modally Transposed",         PASTE_MODAL,                  MAX_PROB/4 },
    { NULL, 0, 0 }
};



MUTATOR paste_blocks =
{
    paste_blocks_func,          /* glues parts of blocks back together */
    paste_blocks_params         /* what sort of paste? */
};



/* stretcher_func:
 *  Stretches a pattern.
 */
PATTERN *stretcher_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c;

    p = clone_pattern(p1);

    if (p) {
	if (i == AUGMENTATION) {
	    TRACE("stretcher_func(AUGMENTATION)\n");

	    for (c=0; c<p->length; c++) {
		p->note[c].time *= 2;
		p->note[c].length *= 2;
	    }
	}
	else {
	    TRACE("stretcher_func(DIMINUTION)\n");

	    for (c=0; c<p->length; c++) {
		p->note[c].time /= 2;
		p->note[c].length /= 2;
	    }
	}
    }

    return p;
}



/* copy_rhythm_func:
 *  Copies rhyhmic material from one pattern to another.
 */
PATTERN *copy_rhythm_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c;
    int x;

    TRACE("copy_rhythm_func()\n");

    p = clone_pattern(p1);

    if (p) {
	x = 0;

	for (c=0; c<p->length; c++) {
	    p->note[c].pitch = p2->note[x].pitch;
	    x++;
	    if (x >= p2->length)
		x = 0;
	}
    }

    return p;
}



/* cut_block_func:
 *  Cuts a small block out of a pattern.
 */
PATTERN *cut_block_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int x1, x2, size, c;

    TRACE("cut_block_func()\n");

    x1 = random() % p1->length;
    x2 = random() % p1->length;

    if (x1 > x2) {
	c = x1;
	x1 = x2;
	x2 = c;
    }

    size = x2 - x1 + 1;

    p = malloc((int)(sizeof(PATTERN) + size*sizeof(NOTE)));

    if (p) {
	p->channel = p1->channel;
	p->length = p->allocated_size = size;

	for (c=0; c<size; c++)
	    p->note[c] = p1->note[x1+c];
    }

    return p;
}



/* paste_blocks_func:
 *  Glues parts of blocks back together. If the i parameter is not
 *  PASTE_AT_PITCH, the second block will be transposed to have the same
 *  median pitch as the first one. This routine randomly decides whether
 *  to paste the entire two patterns together, or just a part of each
 *  one.
 */
PATTERN *paste_blocks_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    NOTE *n;
    long time_gap = 0;
    int pitch_gap;
    int gap1, gap2;
    int c;

    if (i == PASTE_AT_PITCH) {
	TRACE("paste_blocks_func(PASTE_AT_PITCH)\n");
	pitch_gap = 0;
    }
    else {
	pitch_gap = median_pitch(p1) - median_pitch(p2); 

	if (i == PASTE_MODAL) {
	    TRACE("paste_blocks_func(PASTE_MODAL)\n");
	    fill_temp_with_mode(p1);
	}
	else
	    TRACE("paste_blocks_func(PASTE_TRANSPOSED)\n");
    }

    /* skip tail part of pattern 1? */
    if ((PROB(50)) && (p1->length > 1))
	gap1 = random() % (p1->length-1);
    else
	gap1 = 0;

    /* skip first part of pattern 2? */
    if ((PROB(50)) && (p2->length > 1))
	gap2 = random() % (p2->length-1);
    else
	gap2 = 0;

    c = p1->length - gap1 + p2->length - gap2;
    p = malloc((int)(sizeof(PATTERN) + c * sizeof(NOTE)));

    if (p) {
	p->channel = p1->channel;
	p->length = p->allocated_size = c;

	/* copy first pattern */
	for (c=0; c<p1->length-gap1; c++)
	    p->note[c] = p1->note[c];

	/* how big an offset for the times in the second pattern? */
	while (p1->note[p1->length-gap1-1].time < p2->note[gap2].time + time_gap)
	    time_gap -= (CLK_TCK * 60) / metronome_speed;

	while (p1->note[p1->length-gap1-1].time >= p2->note[gap2].time + time_gap)
	    time_gap += (CLK_TCK * 60) / metronome_speed;

	/* copy second pattern */
	for (c=gap2; c<p2->length; c++) {
	    n = p->note + p1->length-gap1+c-gap2;
	    *n = p2->note[c];
	    n->time += time_gap;
	    if (i == PASTE_MODAL)
		n->pitch = closest_note_in_mode(n->pitch + pitch_gap);
	    else
		n->pitch += pitch_gap;
	}
    }

    return p;
}



